/*
    EZLeser - a paystub reader (Einzahlungsschein Leser)

    Copyright (C) 2008  David Mosberger-Tang <dmosberger@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/
#define _GNU_SOURCE
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>

/* Calculate mod10, recursive checksum as specified by the Swiss
   Interbank Clearing's "DTA --- Standards und Formate" document.  */
int
check_mod10 (const char *str)
{
  const char *cp;
  unsigned int carry = 0, digit;
  unsigned char table[10] =
    {
      0, 9, 4, 6, 8, 2, 7, 1, 3, 5
    };

  for (cp = str; *cp; ++cp)
    {
      digit = *cp - '0';
      if (digit > 9)
	return 0;	/* not numeric!! */

      carry = table[(digit + carry) % 10];
    }
  return (carry == 0);	/* for valid sums, carry should be zero at the end */
}

/* Calculate mod11 checksum as specified by the Swiss Interbank
   Clearing's "DTA --- Standards und Formate" document.  */
int
check_mod11 (const char *str)
{
  const char *cp;
  unsigned int digit, sum = 0, rem, weight = 2;

  for (cp = str + strlen (str) - 1; cp >= str + 2; --cp)
    {
      digit = *cp - '0';
      if (digit > 9)
	return 0;
      sum += digit * weight;
      if (++weight > 7)
	weight = 2;
    }
  rem = sum % 11;
  if (rem)
    rem = 11 - rem;

  if (str[0] == '0' + (rem / 10) && str[1] == '0' + (rem % 10))
    return 1;
  return 0;
}

int
main (int argc, char **argv)
{
  char *buf = NULL, *line = NULL, *src, *dst;
  char *refnr = NULL, *acctnr = NULL, *amount = NULL;
  size_t size = 0, acctnr_len = 0, blen;
  ssize_t len, buf_len = 0;
  int numeric = 0;
  FILE *ifp;

  if (argc > 1)
    ifp = fopen (argv[1], "r");
  else
    ifp = stdin;

  while ((len = getline (&line, &size, ifp)) >= 0)
    {
      if (len <= 1)
	continue;
      line[--len] = '\0';

      if (buf_len < len + 1)
	{
	  buf = realloc (buf, len + 1);
	  if (!buf)
	    {
	      perror ("realloc");
	      exit (-1);
	    }
	}

      dst = buf;
      for (src = line; src < line + len; ++src)
	{
	  if (numeric)
	    {
	      if (isdigit (*src))
		*dst++ = *src;
	      else if (!isblank (*src))
		{
		  /* end of numeric string: */
		  *dst++ = '\0';
		  switch (*src)
		    {
		    case '>':
		      blen = strlen (buf);
		      if (!amount && blen >= 13)
			amount = strdup (buf + blen - 13);
		      else
			acctnr = strdup (buf);
		      break;

		    case '+':
		      refnr = strdup (buf);
		      break;

		    default:
		      break;
		    }
		  numeric = 0;
		  dst = buf;
		}
	    }
	  else
	    {
	      numeric = isdigit (*src);
	      if (numeric)
		*dst++ = *src;
	    }
	}
      if (amount && refnr && acctnr)
	{
	  acctnr_len = strlen (acctnr);
	  if (acctnr_len == 9
	      /* Slips with 9-digit account number (ESR9) use mod10: */
	      && check_mod10 (amount)
	      && check_mod10 (refnr) && check_mod10 (acctnr))
	    break;
	  else if (acctnr_len == 5)
	    {
	      /* Slips with 5-digit account number (ESR5) use mod11
		 algorithm over the entire amount/refnr/acctnr string: */
	      size_t amount_len = strlen (amount);
	      size_t refnr_len = strlen (refnr);
	      char *str;
	      int good;

	      str = malloc (amount_len + refnr_len + acctnr_len);
	      if (!str)
		{
		  perror ("malloc");
		  exit (-1);
		}
	      memcpy (str, amount, amount_len);
	      memcpy (str + amount_len, refnr, refnr_len);
	      memcpy (str + amount_len + refnr_len, acctnr, acctnr_len);
	      str[amount_len + refnr_len + acctnr_len - 1] = '\0';
	      good = check_mod11 (str);
	      free (str);
	      if (good)
		break;
	    }
	}
      if (amount)
	{
	  free (amount);
	  amount = NULL;
	}
      if (refnr)
	{
	  free (refnr);
	  refnr = NULL;
	}
      if (acctnr)
	{
	  free (acctnr);
	  acctnr = NULL;
	}
    }

  if (refnr)
    printf ("refnr=%s\n", refnr);
  if (acctnr)
    {
      char *cp;
      int i;

      printf ("acctnr=%s\n", acctnr);
      cp = acctnr;
      printf ("acctnrFormatted=%c", *cp++);
      if (*cp == '1' || *cp == '3')
	printf ("%c", *cp++);
      printf ("-");
      for (i = 0; i < 6; ++i)
	{
	  printf ("%c", *cp++);
	  if (!cp[1])
	    break;
	}
      printf ("-%c\n", *cp);
    }
  if (amount)
    {
      char *src, *dst = amount, *end, *frac;
      size_t l = strlen (amount);
      int zero = 1;

      if (acctnr_len == 9)
	{
	  src = amount + 2;	/* skip over "Belegartcode" (always "01"??) */
	  end = amount + l - 1;	/* drop checksum */
	}
      else
	{
	  /* skip 2 checkdigits, 2 reserve digits, 2 "Belegartcode" digits: */
	  src = amount + 6;
	  end = amount + l;
	}

      while (src < end)
	{
	  int ch = *src++;

	  zero = zero && (ch == '0');
	  if (zero && end - src >= 3)
	    continue;
	  *dst++ = ch;
	  if (end - src == 2)
	    *dst++ = '.';
	}
      *dst++ = '\0';
      printf ("amount=%s\n", amount);

      frac = strchr (amount, '.');
      if (frac)
	*frac++ = '\0';
      else
	frac = "00";
      printf ("amountWhole=%s\n", amount);
      printf ("amountFraction=%s\n", frac);
    }

  return 0;
}
