/*
    ppmautorot - This program is part of EZLeser, a paystub reader

    Copyright (C) 2008  David Mosberger-Tang <dmosberger@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/
#include <math.h>
#include <pgm.h>
#include <ppm.h>
#include <stdio.h>
#include <string.h>

#define MASK_SIZE	23

#define EPS	(3 / 180.0 * M_PI)

struct point
  {
    int col;
    int row;
  };

struct rect
  {
    struct point top_left;
    struct point bot_right;
  };

static char *prog_name;

enum cornerpt
  {
    TL,
    TR,
    BR,
    BL
  };

static const char *corner_name[] = { "TL", "TR", "BR", "BL" };

/* This describes what extreme point (left, top, right, bottom) maps
   to what corner in the rotated image, provided the rotation angle is
   in the range 0 .. M_PI/2.  For angles outside this quadrant, the
   positions need to be shifted.  */
static enum cornerpt map_to_corner[4] =
  {
    TL, TR, BR, BL
  };

static float
square (int val)
{
  return val * val;
}

static void
bound (int *val, int max)
{
  if (*val < 0)
    *val = 0;
  else if (*val > max)
    *val = max;
}

static int
reddish (pixel p)
{
  pixval r, g, b;

  r = PPM_GETR (p);
  g = PPM_GETG (p);
  b = PPM_GETB (p);
  if (r > 0x80 && r > (g + 3) && r > (b + 3))
    return 1;
  if (r < 0x80 && g < 0x80 && b < 0x80)
    return 1;
  return 0;
}

static bit
looks_reddish (pixel **img, int c, int r)
{
  int i, j, w = 0;

  for (i = 0; i < MASK_SIZE; ++i)
    for (j = 0; j < MASK_SIZE; ++j)
      w += reddish (img[r + i][c + j]);

  return (w >= 14 * MASK_SIZE * MASK_SIZE / 16) ? PBM_WHITE : PBM_BLACK;
}

static void
filter (int num_cols, int num_rows, pixel **img, bit **map)
{
  int r, c;

  for (r = 0; r < num_rows - MASK_SIZE; r += MASK_SIZE)
    for (c = 0; c < num_cols - MASK_SIZE; c += MASK_SIZE)
      map[r / MASK_SIZE][c / MASK_SIZE] = looks_reddish (img, c, r);
}

static void
smooth (int num_cols, int num_rows, bit **map)
{
  int r, c;

  for (r = 1; r < num_rows - 1; ++r)
    for (c = 1; c < num_cols - 1; ++c)
      {
	bit ref, pix = map[r][c];

	ref = map[r - 1][c];
	if (map[r + 1][c] == ref)
	  pix = ref;

	ref = map[r][c - 1];
	if (map[r][c + 1] == ref)
	  pix = ref;

	map[r][c] = pix;
      }
}

static float
find_angle_and_bbox (int num_cols, int num_rows, bit **map,
		     struct point *pt, enum cornerpt *pt_loc,
		     struct rect *bbox)
{
  enum extremes_t {
    LFT = 0,
    TOP,
    RGT,
    BOT,
    NUM_EXTREMES
  };
  /* Among all the white pixels, find the four extreme points, whose
     coordinates are, respectively, the min/max row, and min/max
     column.  In case of amiguity, the left-most-point moves towards
     the top, the right-most point moves towards the bottom, the
     top-most point moves towards the right, and the lowest point
     moves towards the left.  That way, if the white pixels form a
     horizontal rectangle, the four points coincide with the corners
     of the rectangle.  */
  struct point e[NUM_EXTREMES], *p1, *p2;
  float a, alpha[NUM_EXTREMES];
  float len, max_len;
  int row, col, i, max_len_i, slides = 0, slide_count[NUM_EXTREMES] = { 0, };
  unsigned outlier = 0;
  int quadrant;

  bbox->top_left.row = num_rows;
  bbox->top_left.col = num_cols;
  bbox->bot_right.row = -1;
  bbox->bot_right.col = -1;
  e[LFT].col = num_cols;
  e[RGT].col = -1;
  e[TOP].row = num_rows;
  e[BOT].row = -1;

  for (row = 0; row < num_rows; ++row)
    for (col = 0; col < num_cols; ++col)
      {
	if (map[row][col] == PBM_WHITE)
	  {
	    if (row > e[BOT].row || (row == e[BOT].row && col < e[BOT].col))
	      {
		if (row == e[BOT].row)
		  ++slide_count[BOT];
		else
		  slide_count[BOT] = 0;
		e[BOT].col = col;
		e[BOT].row = row;
	      }
	    if (row < e[TOP].row || (row == e[TOP].row && col > e[TOP].row))
	      {
		if (row == e[TOP].row)
		  ++slide_count[TOP];
		else
		  slide_count[TOP] = 0;
		e[TOP].col = col;
		e[TOP].row = row;
	      }
	    if (col > e[RGT].col || (col == e[RGT].col && row > e[RGT].row))
	      {
		if (col == e[RGT].col)
		  ++slide_count[RGT];
		else
		  slide_count[RGT] = 0;
		e[RGT].col = col;
		e[RGT].row = row;
	      }
	    if (col < e[LFT].col || (col == e[LFT].col && row < e[LFT].row))
	      {
		if (col == e[LFT].col)
		  ++slide_count[LFT];
		else
		  slide_count[LFT] = 0;
		e[LFT].col = col;
		e[LFT].row = row;
	      }
	    if (row < bbox->top_left.row)
	      bbox->top_left.row = row;
	    if (row > bbox->bot_right.row)
	      bbox->bot_right.row = row;
	    if (col < bbox->top_left.col)
	      bbox->top_left.col = col;
	    if (col > bbox->bot_right.col)
	      bbox->bot_right.col = col;
	  }
      }
  if (e[RGT].col == -1)
    {
      fprintf (stderr, "%s: not a single pixel found!\n", prog_name);
      exit (-1);
    }
  for (i = 0; i < NUM_EXTREMES; ++i)
    slides += slide_count[i];

  if (slides > 10)
    {
      /* Looks like we're dealing with a vertical or horizontal
         rectangle; set the extremes based on the bounding-box, since
         that's more stable than trying to follow along a long, noisy
         edge.  */
      e[LFT].row = bbox->top_left.row;
      e[LFT].col = bbox->top_left.col;
      e[RGT].row = bbox->bot_right.row;
      e[RGT].col = bbox->bot_right.col;
      e[TOP].row = bbox->top_left.row;
      e[TOP].col = bbox->bot_right.col;
      e[BOT].row = bbox->bot_right.row;
      e[BOT].col = bbox->top_left.col;
    }

  fprintf (stderr, "slide_count=%d\n", slides);
  fprintf (stderr, "extremes: l=%d,%d r=%d,%d t=%d,%d b=%d,%d\n",
	   e[LFT].col, e[LFT].row, e[RGT].col, e[RGT].row,
	   e[TOP].col, e[TOP].row, e[BOT].col, e[BOT].row);
  fprintf (stderr, "bbox: tl=%u,%u pt=%u,%u\n",
	   bbox->top_left.col, bbox->top_left.row,
	   bbox->bot_right.col, bbox->bot_right.row);

  alpha[0] = atan2f (e[LFT].row - e[TOP].row, e[TOP].col - e[LFT].col);
  alpha[1] = atan2f (e[RGT].col - e[TOP].col, e[RGT].row - e[TOP].row);
  alpha[2] = atan2f (e[BOT].row - e[RGT].row, e[RGT].col - e[BOT].col);
  alpha[3] = atan2f (e[BOT].col - e[LFT].col, e[BOT].row - e[LFT].row);

  for (i = 0; i < NUM_EXTREMES; ++i)
    fprintf (stderr, "alpha[%d]=%9.3f\n", i, alpha[i] * 180 / M_PI);

  if (fabsf (alpha[0] - alpha[1]) < EPS || fabs (alpha[0] - alpha[2]) < EPS)
    a = alpha[0];
  else
    a = alpha[1];

  for (i = 0; i < NUM_EXTREMES; ++i)
    if (alpha[i] - a >= EPS)
      outlier |= (1u << i);

  max_len = 0;
  max_len_i = 0;
  for (i = 0; i < NUM_EXTREMES; ++i)
    {
      if (outlier & (1u << i))
	continue;
      p1 = &e[i];
      p2 = &e[(i + 1) % NUM_EXTREMES];
      len = sqrt (square (p2->col - p1->col) + square (p2->row - p1->row));
      if (len > max_len)
	{
	  max_len = len;
	  max_len_i = i;
	}
    }
  fprintf (stderr, "max_len_i=%d max_len=%9.1f (outlier=%x)\n",
	   max_len_i, max_len, outlier);

  if (max_len_i == 1)
    a -= M_PI / 2;
  else if (max_len_i == 2)
    a += M_PI;
  else if (max_len_i == 3)
    a += M_PI / 2;

  if (a >= M_PI)
    quadrant = 2;
  else if (a >= M_PI / 2)
    quadrant = 1;
  else if (a >= 0)
    quadrant = 0;
  else if (a >= -M_PI / 2 - 1e-5)
    quadrant = 3;
  else
    quadrant = 2;
  *pt_loc = map_to_corner[(max_len_i + quadrant) % 4];
  *pt = e[max_len_i];

  return a;
}

static pixval
get_threshold (int num_cols, int num_rows, pixel **img, struct rect *area)
{
  pixval min_lum = 0xffff, max_lum = 0, lum;
  int r, c;

  for (r = area->top_left.row; r <= area->bot_right.row; ++r)
    for (c = area->top_left.col; c <= area->bot_right.col; ++c)
      {
	lum = PPM_LUMIN (img[r][c]);
	if (lum < min_lum)
	  min_lum = lum;
	if (lum > max_lum)
	  max_lum = lum;
      }
  return (min_lum + max_lum) / 2;
}

static void
rotate_and_threshold (int num_cols, int num_rows, bit **map,
		      struct point *pt, enum cornerpt corner,
		      int num_img_cols, int num_img_rows, pixel **img,
		      float angle, pixval threshold)
{
  float img_r0, img_c0;
  float sin_alpha = sinf (angle);
  float cos_alpha = cosf (angle);
  pixval lum;
  int i, j, r, c, r0, c0, rD, cD, img_r, img_c;

  switch (corner)
    {
    case TL: r0 = 0; c0 = 0; rD = 1; cD = 1; break;
    case TR: r0 = 0; c0 = num_cols - 1; rD = 1; cD = -1; break;
    case BR: r0 = num_rows - 1; c0 = num_cols - 1; rD = -1; cD = -1; break;
    case BL: r0 = num_rows - 1; c0 = 0; rD = -1; cD = 1; break;
    }

  r = r0;
  for (i = 0; i < num_rows; ++i, r += rD)
    {
      img_c0 = pt->col + cD * i * sin_alpha;
      img_r0 = pt->row + rD * i * cos_alpha;
      c = c0;
      for (j = 0; j < num_cols; ++j, c += cD)
	{
	  img_c = img_c0 + cD * j * cos_alpha;
	  img_r = img_r0 - rD * j * sin_alpha;
	  if (img_c < 0 || img_c >= num_img_cols
	      || img_r < 0 || img_r >= num_img_rows)
	    lum = threshold + 1;
	  else
	    lum = PPM_LUMIN (img[img_r][img_c]);
	  map[r][c] = (lum > threshold) ? PBM_WHITE : PBM_BLACK;
	}
    }
}

int
main (int argc, char **argv)
{
  int num_cols, num_rows, num_map_rows, num_map_cols, w_bb, h_bb;
  float alpha, denom, sin_alpha, cos_alpha;
  FILE *ifp = stdin, *ofp = stdout;
  pixval max_val, threshold;
  enum cornerpt corner;
  struct rect bbox;
  struct point pt;
  pixel **img;
  bit **map;

  prog_name = strrchr (argv[0], '/');
  if (prog_name)
    ++prog_name;
  else
    prog_name = argv[0];

  ppm_init (&argc, argv);

  if (argc >= 2)
    {
      ifp = fopen (argv[1], "r");
      if (!ifp)
	{
	  perror (argv[1]);
	  exit (-1);
	}
      if (argc >= 3)
	{
	  ofp = fopen (argv[2], "w");
	  if (!ofp)
	    {
	      perror (argv[2]);
	      exit (-1);
	    }
	}
    }

  img = ppm_readppm (ifp, &num_cols, &num_rows, &max_val);
  if (!img)
    {
      fprintf (stderr, "Failed to read image\n");
      exit (-1);
    }

  num_map_rows = num_rows / MASK_SIZE;
  num_map_cols = num_cols / MASK_SIZE;

  map = pbm_allocarray (num_map_cols, num_map_rows);
  filter (num_cols, num_rows, img, map);
  smooth (num_map_cols, num_map_rows, map);

#if 0
  FILE *tfp = fopen ("map.pbm", "w");
  if (tfp)
    {
      pbm_writepbm (tfp, map, num_map_cols, num_map_rows, 0);
      fclose (tfp);
      tfp = NULL;
    }
#endif

  alpha = find_angle_and_bbox (num_map_cols, num_map_rows, map,
			       &pt, &corner, &bbox);
  fprintf (stderr, "rotation angle = %9.3f degrees\n", alpha / M_PI * 180.);

  pt.row = (pt.row) * MASK_SIZE;
  pt.col = (pt.col) * MASK_SIZE;
  bbox.top_left.row = (bbox.top_left.row - 6) * MASK_SIZE;
  bbox.top_left.col = (bbox.top_left.col - 6) * MASK_SIZE;
  bbox.bot_right.row = (bbox.bot_right.row + 6) * MASK_SIZE;
  bbox.bot_right.col = (bbox.bot_right.col + 6) * MASK_SIZE;
  bound (&bbox.top_left.row, num_rows - 1);
  bound (&bbox.top_left.col, num_cols - 1);
  bound (&bbox.bot_right.row, num_rows - 1);
  bound (&bbox.bot_right.col, num_cols - 1);

  fprintf (stderr, "bbox: tl=%u,%u br=%u,%u %s: %u,%u\n",
	   bbox.top_left.col, bbox.top_left.row,
	   bbox.bot_right.col, bbox.bot_right.row,
	   corner_name[corner], pt.col, pt.row);

  pbm_freearray (map, num_map_rows);

  w_bb = bbox.bot_right.col - bbox.top_left.col + 1;
  h_bb = bbox.bot_right.row - bbox.top_left.row + 1;
  sin_alpha = sinf (fabsf (alpha));
  cos_alpha = cosf (fabsf (alpha));
  denom = sin_alpha * sin_alpha - cos_alpha * cos_alpha;
  num_map_cols = (h_bb * sin_alpha - w_bb * cos_alpha) / denom;
  num_map_rows = (w_bb * sin_alpha - h_bb * cos_alpha) / denom;
  fprintf (stderr, "w=%d h=%d\n", num_map_cols, num_map_rows);
  map = pbm_allocarray (num_map_cols, num_map_rows);

  threshold = get_threshold (num_cols, num_rows, img, &bbox);
  fprintf (stderr, "threshold = %d\n", threshold);

  rotate_and_threshold (num_map_cols, num_map_rows, map, &pt, corner,
			num_cols, num_rows, img,
			alpha, threshold);

  pbm_writepbm (ofp, map, num_map_cols, num_map_rows, 0);

  return 0;
}
